import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)

/**
 * Write a description of class Worldy here.
 * 
 * @author (your name) 
 * @version (a version number or a date)
 */
public class Worldy extends World {
    int x = 0;
    int y = 0;
    double minX;
    double maxX;
    double minY;
    double maxY;
    double pixelSizeX;
    double pixelSizeY;
    Complex cons = new Complex(0, 0);
    boolean mandelbrot = true;
    int preX = -1;
    int preY = -1;
    Color[] colors = new Color[1000];
    GreenfootImage fractal;
    GreenfootImage line;
    
    public Worldy() {
        super(600, 400, 1);
        reset(-3.5, -2, 2.5, 2);
        for(int i = 0; i < colors.length; i++) {
            colors[i] = new Color(126+Greenfoot.getRandomNumber(126), 126+Greenfoot.getRandomNumber(126), 126+Greenfoot.getRandomNumber(126));
        }
        Greenfoot.start();
    }
    
    public void reset(double minX, double minY, double maxX, double maxY) {
        fractal = new GreenfootImage(getWidth(), getHeight());
        line = new GreenfootImage(getWidth(), getHeight());
        this.minX = minX;
        this.minY = minY;
        this.maxX = maxX;
        this.maxY = maxY;
        x = 0;
        y = 0;
        pixelSizeX = (maxX-minX)/getWidth();
        pixelSizeY = (maxY-minY)/getHeight();
    }
    
    public int mandel(double x, double y, int iter) {
        Complex z = new Complex(0, 0);
        Complex c = new Complex(x, y);
        for(int i = 1; i <= iter; i++) {
            double zr = (z.r*z.r-z.i*z.i)+c.r;
            z.i = 2*z.r*z.i+c.i;
            z.r = zr;
            if(z.r*z.r+z.i*z.i > 4) return (int)(z.r*z.r+z.i*z.i);
        }
        return 0;
    }
    
    public int deadAnimal(double x, double y, int iter) {
        Complex z = new Complex(0, 0);
        Complex c = new Complex(x, y);
        for(int i = 1; i <= iter; i++) {
            double zr = (Math.pow(z.r, 4)-3*z.r*z.r*z.i*z.i+Math.pow(z.i, 4))+c.r;
            z.i = (4*z.r*z.r*z.r*z.i-3*z.r*z.r*z.i*z.i*z.i-z.r*z.i*z.i*z.i)+c.i;
            z.r = zr;
            if(z.r*z.r+z.i*z.i >= 4) return (int)(z.r*z.r+z.i*z.i);
        }
        return 0;
    }
    
    public int julia(double x, double y, int iter) {
        Complex z = new Complex(x, y);
        for(int i = 1; i <= iter; i++) {
            double zr = (z.r*z.r-z.i*z.i)+cons.r;
            z.i = 2*z.r*z.i+cons.i;
            z.r = zr;
            if(z.r*z.r+z.i*z.i > 4) return (int)(z.r*z.r+z.i*z.i);
        }
        return 0;
    }
    
    public void drawTrail(Complex pos, int iter) {
        line.clear();
        line.setColor(Color.WHITE);
        if(!mandelbrot) {
            Complex z = pos;
            for(int i = 1; i <= iter; i++) {
                Complex preZ = new Complex(z.r, z.i);
                z.i = 2*preZ.r*preZ.i+cons.i;
                z.r = (preZ.r*preZ.r-preZ.i*preZ.i)+cons.r;
                line.drawLine((int)((preZ.r-minX)/pixelSizeX), (int)((preZ.i-minY)/pixelSizeY), (int)((z.r-minX)/pixelSizeX), (int)((z.i-minY)/pixelSizeY));
            }
        }
    }
    
    public void act() {
        if(y < getHeight()) {
            double imag = minY + y*pixelSizeY;
            if(x < getWidth()) {
                double real = minX + x*pixelSizeX;
                int iter = 256;
                int color = 0;
                if(mandelbrot) {
                    color = mandel(real, imag, iter);
                } else {
                    color = julia(real, imag, iter);
                }
                
                double colorFrac = color;
                colorFrac /= 4;
                if(color > 0) fractal.setColorAt(x, y, colors[(int)(colorFrac%colors.length)]);
                else fractal.setColorAt(x, y, Color.BLACK);
                x++;
            } else {
                x = 0;
                y++;
            }
        }
        MouseInfo mouse = Greenfoot.getMouseInfo();
        if(mouse != null) {
            if(Greenfoot.mouseClicked(null)) {
                double realMin = -3;
                double imagMin = -2;
                double realMax = 3;
                double imagMax = 2;
                if(Greenfoot.isKeyDown("control")) {
                    mandelbrot = !mandelbrot;
                    if(mandelbrot) {
                        realMin = -3.5;
                        realMax = 2.5;
                    }
                    cons = new Complex(minX + mouse.getX()*pixelSizeX, minY + mouse.getY()*pixelSizeY);
                } else if(Greenfoot.isKeyDown("shift")) {
                    realMin = minX + (mouse.getX()-getWidth())*pixelSizeX;
                    imagMin = minY + (mouse.getY()-getHeight())*pixelSizeY;
                    realMax = minX + (mouse.getX()+getWidth())*pixelSizeX;
                    imagMax = minY + (mouse.getY()+getHeight())*pixelSizeY;
                } else {
                    realMin = minX + (mouse.getX()-getWidth()/4)*pixelSizeX;
                    imagMin = minY + (mouse.getY()-getHeight()/4)*pixelSizeY;
                    realMax = minX + (mouse.getX()+getWidth()/4)*pixelSizeX;
                    imagMax = minY + (mouse.getY()+getHeight()/4)*pixelSizeY;
                }
                reset(realMin, imagMin, realMax, imagMax);
            } else if("space".equals(Greenfoot.getKey())) {
                double realMin = -3;
                double imagMin = -2;
                double realMax = 3;
                double imagMax = 2;
                if(mandelbrot) {
                    realMin = -3.5;
                    realMax = 2.5;
                }
                reset(realMin, imagMin, realMax, imagMax);
            } else {
                if(mouse.getX()!=preX || mouse.getY()!=preY) {
                    Complex pos = new Complex(minX + mouse.getX()*pixelSizeX, minY + mouse.getY()*pixelSizeY);
                    int iter = 100;
                    drawTrail(pos, iter);
                    preX = mouse.getX();
                    preY = mouse.getY();
                }
            }
            if(y >= getHeight()) {
                getBackground().drawImage(fractal, 0, 0);
                if(!mandelbrot && Greenfoot.isKeyDown("z")) getBackground().drawImage(line, 0, 0);
            }
        }
    }
}
